/**
 * Extension that allows authors to display context tooltips bound to specific
 * positions
 */
"use strict";

import tween from '../vendor/tween';
import {extend, prefixed, makePos, has3d} from '../utils';
import * as dom from '../dom';

var instance = null;
var lastTween = null;

export var alignDefaults = {
	/** CSS selector for getting popup tail */
	tailClass: 'CodeMirror-tooltip__tail',

	/** Class name for switching tail/popup position relative to target point */
	belowClass: 'CodeMirror-tooltip_below',

	/** Min distance between popup and viewport */
	popupMargin: 5,

	/** Min distance between popup left/right edge and its tail */
	tailMargin: 11
};

export var actions = {
	/**
	 * Shows tooltip with given text, wait for `options.wait`
	 * milliseconds then hides tooltip
	 */
	tooltip({ options, editor, timer, next, prev, last }) {
		options = extend({
			wait: 0,
			// pos: 'caret'
			pos: null
		}, wrap('text', options));

		if (!options.pos) {
			throw new Error('tooltip `pos` is required');
		}

		var pos = makePos(options.pos, editor);
		editor.setCursor(pos);

		pos = fixCaretPos(editor.charCoords(pos));

		show(options.text, {
			pos, prev, next,
			wait: options.wait,
			timer, editor, last
		}, () => {});
	}
};

export function show(text, { pos, prev, next, timer, wait, editor, last }, callback) {
	hide();

	instance = dom.toDOM(`<div class="CodeMirror-tooltip">
		<div class="CodeMirror-tooltip__content">
			<div class="CodeMirror-tooltip__text">
				${text}
			</div>
			<div class="CodeMirror-tooltip__buttons">
				${!wait && prev ?
					'<button type="button" class="CodeMirror-tooltip__prev">back</button>'
				: ''}
				${!wait && next ?
					`<button type="button" class="CodeMirror-tooltip__next">${
						last ? 'finish' : 'next'
					}</button>`
				: ''}
			</div>
		</div>
		<div class="CodeMirror-tooltip__tail"></div>
		</div>`);

	const prevButton = instance
		.querySelector('.CodeMirror-tooltip__prev');
	const nextButton = instance
		.querySelector('.CodeMirror-tooltip__next');

	if (prevButton) {
		prevButton.onclick = () => hide(() => timer(prev));
	}

	if (nextButton) {
		nextButton.onclick = () => hide(() => timer(next));
	}

	if (wait) {
		timer(() => hide(() => timer(next)), wait);
	}

	dom.css(instance, prefixed('transform'), 'scale(0)');
	document.body.appendChild(instance);

	alignPopupWithTail(instance, {position: pos});
	animateShow(instance, {complete: callback});
}

export function hide(callback) {
	if (instance) {
		if (lastTween) {
			lastTween.stop();
			lastTween = null;
		}

		const prevButton = instance
			.querySelector('.CodeMirror-tooltip__prev');
		const nextButton = instance
			.querySelector('.CodeMirror-tooltip__next');

		if (prevButton) {
			prevButton.onclick = null;
		}

		if (nextButton) {
			nextButton.onclick = null;
		}

		animateHide(instance, {complete: callback});

		instance = null;
	} else if (callback) {
		callback();
	}
}

/**
 * Helper function that finds optimal position of tooltip popup on page
 * and aligns popup tail with this position
 * @param {Element} popup
 * @param {Object} options
 */
function alignPopupWithTail(popup, options={}) {
	options = extend({}, alignDefaults, options);

	dom.css(popup, {
		left: 0,
		top: 0
	});

	var tail = popup.querySelector('.' + options.tailClass);

	var resultX = 0, resultY = 0;
	var pos = options.position;
	var vp = dom.viewportRect();

	var width = popup.offsetWidth;
	var height = popup.offsetHeight;

	var isTop;

	// calculate horizontal position
	resultX = Math.min(vp.width - width - options.popupMargin, Math.max(options.popupMargin, pos.x - vp.left - width / 2));

	// calculate vertical position
	if (height + tail.offsetHeight + options.popupMargin + vp.top < pos.y) {
		// place above target position
		resultY = Math.max(0, pos.y - height - tail.offsetHeight);
		isTop = true;
	} else {
		// place below target position
		resultY = pos.y + tail.offsetHeight + pos.height;
		isTop = false;
	}

	// calculate tail position
	var tailMinLeft = options.tailMargin;
	var tailMaxLeft = width - options.tailMargin;
	tail.style.left = Math.min(tailMaxLeft, Math.max(tailMinLeft, pos.x - resultX - vp.left)) + 'px';

	dom.css(popup, {
		left: resultX,
		top: resultY
	});

	popup.classList.toggle(options.belowClass, !isTop);
}

/**
 * @param {jQuery} elem
 * @param {Object} options
 */
function animateShow(elem, options={}) {
	options = extend({}, alignDefaults, options);
	var cssOrigin = prefixed('transform-origin');
	var cssTransform = prefixed('transform');
	var style = elem.style;

	var tail = elem.querySelector('.' + options.tailClass);
	var xOrigin = dom.css(tail, 'left');
	var yOrigin = tail.offsetTop;
	if (elem.classList.contains(options.belowClass)) {
		yOrigin -= tail.offsetHeight;
	}

	yOrigin += 'px';

	style[cssOrigin] = xOrigin + ' ' + yOrigin;
	var prefix = has3d ? 'translateZ(0) ' : '';

	return lastTween = tween({
		duration: 300,
		easing: 'outBack',
		step(pos) {
			style[cssTransform] = prefix + 'scale(' + pos + ')';
		},
		complete() {
			style[cssTransform] = 'none';
			lastTween = null;
			options.complete && options.complete(elem);
		}
	});
}

/**
 * @param {jQuery} elem
 * @param {Object} options
 */
function animateHide(elem, options) {
	var style = elem.style;

	return tween({
		duration: 200,
		easing: 'linear',
		step: function(pos) {
			style.opacity = (1 - pos);
		},
		complete: function() {
			dom.remove(elem);
			options.complete && options.complete(elem);
		}
	});
}

/**
 * Resolves position where tooltip should point to
 * @param {Object} pos
 * @param {CodeMirror} editor
 * @returns {Object} Object with <code>x</code> and <code>y</code>
 * properties
 */
/*function resolvePosition(pos, editor) {
	if (pos === 'caret') {
		// get absolute position of current caret position
		return fixCaretPos(editor.cursorCoords(true));
	}

	if (typeof pos === 'object') {
		if ('x' in pos && 'y' in pos) {
			// passed absolute coordinates
			return pos;
		}

		if ('left' in pos && 'top' in pos) {
			// passed absolute coordinates
			return fixCaretPos(pos);
		}
	}

	pos = makePos(pos, editor);
	return fixCaretPos(editor.charCoords(pos));
}*/

function fixCaretPos(pos) {
	if ('left' in pos) {
		pos.x = pos.left;
	}

	if ('top' in pos) {
		pos.y = pos.top;
	}

	if ('bottom' in pos) {
		pos.height = pos.bottom - pos.top;
	} else {
		pos.height = 0;
	}

	return pos;
}

function wrap(key, value) {
	return typeof value === 'object' ? value : {[key]: value};
}